--
-- pgcryptokey
--

CREATE TABLE pgcryptokey (
	key_id SERIAL PRIMARY KEY,
	name TEXT DEFAULT 'main',
	encrypted_data_key BYTEA NOT NULL,
	created TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP NOT NULL,
	superseded_by INTEGER
);

-- Allow only one null superseded_by row per key name
CREATE UNIQUE INDEX i_pgcryptokey ON pgcryptokey (name) WHERE superseded_by IS NULL;

--
-- Get shared key
--

CREATE FUNCTION get_shared_key()
RETURNS TEXT
AS $$
DECLARE
	shared_key TEXT;
	access_password TEXT;
	addr TEXT;
	ssl TEXT;
BEGIN
	-- shared_key already set?
	SELECT current_setting('pgcryptokey.shared_key', true) INTO shared_key;
	IF NOT FOUND OR shared_key IS NOT NULL THEN
	    RAISE EXCEPTION 'pgcryptokey.shared_key is already set;  use current_setting() to get its value';
	END IF;

	-- Must have a secure connection to safely send this value, so
	-- must be using SSL or Unix-domain sockets.
	SELECT inet_client_addr() INTO addr;
	SELECT current_setting('ssl', true) INTO ssl;
	IF NOT FOUND OR ((ssl IS NULL OR ssl != 'on') AND (addr IS NOT NULL OR length(addr) != 0)) THEN
	    RAISE EXCEPTION 'function can only be called when in SSL mode or when using Unix-domain sockets';
	END IF;

	-- the hex of the random bytes is used as the shared key
	SELECT set_config('pgcryptokey.shared_key',
			  encode(gen_random_bytes(32), 'hex'),
			  FALSE) INTO shared_key;

	RETURN shared_key;
END
$$ LANGUAGE plpgsql;

--
-- Securely set access password
--

CREATE FUNCTION set_session_access_password(encrypted_password TEXT)
RETURNS BOOLEAN
AS $$
DECLARE
	shared_key TEXT;
	access_password TEXT;
BEGIN
	-- Is shared_key set?
	SELECT current_setting('pgcryptokey.shared_key', true) INTO shared_key;
	IF NOT FOUND OR shared_key IS NULL OR length(shared_key) = 0 THEN
	    RAISE EXCEPTION 'pgcryptokey.shared_key is not set;  call get_shared_key()';
	END IF;

	IF (encrypted_password IS NULL OR length(encrypted_password) = 0) THEN
	    RAISE EXCEPTION 'invalid encrypted password';
	END IF;

	-- validate encrypted_password is hex
	IF btrim(encrypted_password) !~ '^[0-9a-fA-F]*$' OR length(btrim(encrypted_password)) % 2 != 0 THEN
	    RAISE EXCEPTION 'the access password must be encrypted by the shared key, and in hex format';
	END IF;

	-- decrypt encrypted_password to get the access password, convert to hex, and set the GUC
	-- We effectively use the hex of the access password that the client encrypted
	-- because the pgp_sym_encrypt_bytea/pgp_sym_decrypt_bytea password must be TEXT.
	SELECT set_config('pgcryptokey.access_password',
			  encode(pgp_sym_decrypt_bytea(decode(encrypted_password, 'hex'),
						       shared_key),
				 'hex'),
			  FALSE) INTO access_password;
	IF NOT FOUND OR length(access_password) = 0 THEN
	    RAISE EXCEPTION 'cannot set access password';
	END IF;

	RETURN TRUE;
END
$$ LANGUAGE plpgsql;

--
-- Change access password
--

CREATE FUNCTION change_key_access_password(v_key_id INTEGER, new_encrypted_password TEXT)
RETURNS BOOLEAN
AS $$
DECLARE
	shared_key TEXT;
	old_access_password TEXT;
BEGIN
	-- Is shared_key set?
	SELECT current_setting('pgcryptokey.shared_key', true) INTO shared_key;
	IF NOT FOUND OR shared_key IS NULL OR length(shared_key) = 0 THEN
	    RAISE EXCEPTION 'pgcryptokey.shared_key is not set;  call get_shared_key()';
	END IF;

	-- Is the access password set?
	SELECT current_setting('pgcryptokey.access_password', true) INTO old_access_password;
	IF NOT FOUND OR old_access_password IS NULL OR length(old_access_password) = 0 THEN
	    RAISE EXCEPTION 'pgcryptokey.access_password is not set;  call set_session_access_password()';
	END IF;

	IF (new_encrypted_password IS NULL OR length(new_encrypted_password) = 0) THEN
	    RAISE EXCEPTION 'invalid new access password';
	END IF;

	-- validate new_encrypted_password is hex
	IF btrim(new_encrypted_password) !~ '^[0-9a-fA-F]*$' OR length(btrim(new_encrypted_password)) % 2 != 0 THEN
	    RAISE EXCEPTION 'the new access password must be encrypted by the shared key, and in hex format';
	END IF;

	-- old access_password is checked here
	UPDATE pgcryptokey SET encrypted_data_key =
		pgp_sym_encrypt_bytea(pgp_sym_decrypt_bytea(pgcryptokey.encrypted_data_key,
							    old_access_password),
				      encode(pgp_sym_decrypt_bytea(decode(new_encrypted_password, 'hex'),
								   shared_key),
					     'hex'))
	WHERE pgcryptokey.key_id = v_key_id;
	IF NOT FOUND THEN
	    RAISE EXCEPTION 'cryptokey entry not updated';
	END IF;

	RETURN TRUE;
END
$$ LANGUAGE plpgsql;

CREATE FUNCTION change_key_access_password(v_name TEXT, new_encrypted_password TEXT)
RETURNS BOOLEAN
AS $$
DECLARE
	v_key_id INTEGER;
BEGIN
	SELECT key_id INTO v_key_id
	FROM pgcryptokey
	WHERE pgcryptokey.name = v_name AND superseded_by IS NULL;
	IF NOT FOUND THEN
	    RAISE EXCEPTION 'cryptokey entry not found';
	END IF;
	RETURN change_access_password(v_key_id, new_encrypted_password);
END
$$ LANGUAGE plpgsql;

--
-- Create cryptokey
--

CREATE FUNCTION create_cryptokey(name TEXT, byte_len INTEGER)
RETURNS INTEGER
AS $$
DECLARE
	access_password TEXT;
	v_key_id INTEGER;
BEGIN
	-- Is the access password set?
	SELECT current_setting('pgcryptokey.access_password', true) INTO access_password;
	IF NOT FOUND OR access_password IS NULL OR length(access_password) = 0 THEN
	    RAISE EXCEPTION 'pgcryptokey.access_password is not set;  call set_session_access_password()';
	END IF;

	INSERT INTO pgcryptokey VALUES (DEFAULT, name,
		pgp_sym_encrypt_bytea(gen_random_bytes(byte_len), access_password), DEFAULT, NULL)
	RETURNING key_id INTO v_key_id;

	RETURN v_key_id;
END
$$ LANGUAGE plpgsql;

--
-- Supercede cryptokey
--

CREATE FUNCTION supersede_cryptokey(old_key_id INTEGER, byte_len INTEGER)
RETURNS INTEGER
AS $$
DECLARE
	access_password TEXT;
	v_name TEXT;
	seq_name TEXT;
	new_key_id INTEGER;
BEGIN
	-- Is the access password set?
	SELECT current_setting('pgcryptokey.access_password', true) INTO access_password;
	IF NOT FOUND OR access_password IS NULL OR length(access_password) = 0 THEN
	    RAISE EXCEPTION 'pgcryptokey.access_password is not set;  call set_session_access_password()';
	END IF;

	SELECT name INTO v_name
	FROM pgcryptokey
	WHERE pgcryptokey.key_id = old_key_id;
	IF NOT FOUND THEN
	    RAISE EXCEPTION 'cryptokey entry not found';
	END IF;

	-- get next key_id first so we can do the UPDATE first
	SELECT pg_get_serial_sequence('pgcryptokey', 'key_id') INTO seq_name;
	IF NOT FOUND THEN
	    RAISE EXCEPTION 'sequence name not found';
	END IF;
	SELECT nextval(seq_name) INTO new_key_id;
	IF NOT FOUND THEN
	    RAISE EXCEPTION 'sequence value not found';
	END IF;

	-- must update old row first because of the index
	UPDATE pgcryptokey SET superseded_by = new_key_id
	WHERE key_id = old_key_id AND superseded_by IS NULL;
	IF NOT FOUND THEN
	    RAISE EXCEPTION 'active cryptokey entry not found';
	END IF;

	INSERT INTO pgcryptokey VALUES (new_key_id, v_name,
		pgp_sym_encrypt_bytea(gen_random_bytes(byte_len), access_password), DEFAULT, NULL);

	RETURN new_key_id;
END
$$ LANGUAGE plpgsql;

CREATE FUNCTION supersede_cryptokey(v_name TEXT, byte_len INTEGER)
RETURNS INTEGER
AS $$
DECLARE
	v_key_id INTEGER;
BEGIN
	SELECT key_id INTO v_key_id
	FROM pgcryptokey
	WHERE pgcryptokey.name = v_name AND superseded_by IS NULL;
	IF NOT FOUND THEN
	    RAISE EXCEPTION 'active cryptokey entry not found';
	END IF;

	RETURN supersede_cryptokey(v_key_id, byte_len);
END
$$ LANGUAGE plpgsql;

--
-- Store key in server variable
--

CREATE FUNCTION set_cryptokey(v_key_id INTEGER)
RETURNS BOOLEAN
AS $$
DECLARE
	access_password TEXT;
	key TEXT;
	v_name TEXT;
BEGIN
	-- Is the access password set?
	SELECT current_setting('pgcryptokey.access_password', true) INTO access_password;
	IF NOT FOUND OR access_password IS NULL OR length(access_password) = 0 THEN
	    RAISE EXCEPTION 'pgcryptokey.access_password is not set;  call set_session_access_password()';
	END IF;

	SELECT name, encode(pgp_sym_decrypt_bytea(pgcryptokey.encrypted_data_key, access_password), 'hex')
	INTO v_name, key
	FROM pgcryptokey
	WHERE pgcryptokey.key_id = v_key_id;
	IF NOT FOUND THEN
	    RAISE EXCEPTION 'cryptokey entry not found';
	END IF;

	PERFORM set_config('pgcryptokey.key_id', v_key_id::TEXT, false);
	PERFORM set_config('pgcryptokey.name', v_name, false);
	-- We can only store text in server variables, not bytea, and
	-- pgp_sym_* functions can only take text as a password, so we
	-- will use the hex representation of the bytea value as the
	-- password.
	PERFORM set_config('pgcryptokey.key', key, false);

	RETURN TRUE;
END
$$ LANGUAGE plpgsql;

CREATE FUNCTION set_cryptokey(v_name TEXT)
RETURNS BOOLEAN
AS $$
DECLARE
	v_key_id INTEGER;
BEGIN
	SELECT key_id INTO v_key_id
	FROM pgcryptokey
	WHERE pgcryptokey.name = v_name AND superseded_by IS NULL;
	IF NOT FOUND THEN
	    RAISE EXCEPTION 'cryptokey entry not found';
	END IF;

	RETURN set_cryptokey(v_key_id);
END
$$ LANGUAGE plpgsql;

--
-- Get cryptokey
--

CREATE FUNCTION get_cryptokey(v_key_id INTEGER)
RETURNS TEXT
AS $$
DECLARE
	access_password TEXT;
	key TEXT;
BEGIN
	-- Is the access password set?
	SELECT current_setting('pgcryptokey.access_password', true) INTO access_password;
	IF NOT FOUND OR access_password IS NULL OR length(access_password) = 0 THEN
	    RAISE EXCEPTION 'pgcryptokey.access_password is not set;  call set_session_access_password()';
	END IF;

	SELECT encode(pgp_sym_decrypt_bytea(pgcryptokey.encrypted_data_key, access_password), 'hex')
	INTO key
	FROM pgcryptokey
	WHERE pgcryptokey.key_id = v_key_id;
	IF NOT FOUND THEN
	    RAISE EXCEPTION 'cryptokey entry not found';
	END IF;

	RETURN key;
END
$$ LANGUAGE plpgsql PARALLEL SAFE;

CREATE FUNCTION get_cryptokey(v_name TEXT)
RETURNS TEXT
AS $$
DECLARE
	v_key_id INTEGER;
BEGIN
	SELECT key_id INTO v_key_id
	FROM pgcryptokey
	WHERE pgcryptokey.name = v_name AND superseded_by IS NULL;
	IF NOT FOUND THEN
	    RAISE EXCEPTION 'cryptokey entry not found';
	END IF;

	RETURN get_cryptokey(v_key_id);
END
$$ LANGUAGE plpgsql PARALLEL SAFE;

--
-- Reencrypt data
--

-- we do not allow key names here since it could be confusing for superseded keys
CREATE FUNCTION reencrypt_data(data BYTEA, old_key_id INTEGER, new_key_id INTEGER)
RETURNS BYTEA
AS $$
DECLARE
	access_password TEXT;
BEGIN
	-- Is the access password set?
	SELECT current_setting('pgcryptokey.access_password', true) INTO access_password;
	IF NOT FOUND OR access_password IS NULL OR length(access_password) = 0 THEN
	    RAISE EXCEPTION 'pgcryptokey.access_password is not set;  call set_session_access_password()';
	END IF;

	RETURN pgp_sym_encrypt(
		pgp_sym_decrypt(data, get_cryptokey(old_key_id)), get_cryptokey(new_key_id));
END
$$ LANGUAGE plpgsql PARALLEL SAFE;

CREATE FUNCTION reencrypt_data_bytea(data BYTEA, old_key_id INTEGER, new_key_id INTEGER)
RETURNS BYTEA
AS $$
DECLARE
	access_password TEXT;
BEGIN
	-- Is the access password set?
	SELECT current_setting('pgcryptokey.access_password', true) INTO access_password;
	IF NOT FOUND OR access_password IS NULL OR length(access_password) = 0 THEN
	    RAISE EXCEPTION 'pgcryptokey.access_password is not set;  call set_session_access_password()';
	END IF;

	RETURN pgp_sym_encrypt_bytea(
		pgp_sym_decrypt_bytea(data, get_cryptokey(old_key_id)), get_cryptokey(new_key_id));
END
$$ LANGUAGE plpgsql PARALLEL SAFE;

--
-- Drop cryptokey
--

CREATE FUNCTION drop_cryptokey(v_key_id INTEGER)
RETURNS BOOLEAN
AS $$
BEGIN
	-- verify the access password
	PERFORM set_cryptokey(v_key_id);

	DELETE FROM pgcryptokey
	WHERE pgcryptokey.key_id = v_key_id;
	IF NOT FOUND THEN
	    RAISE EXCEPTION 'cryptokey entry not found';
	END IF;

	RETURN TRUE;
END
$$ LANGUAGE plpgsql;

-- this only deletes the active key, but warns if other matching names exist
CREATE FUNCTION drop_cryptokey(v_name TEXT)
RETURNS BOOLEAN
AS $$
DECLARE
	v_key_id INTEGER;
BEGIN
	-- notify if superseded keys exist
	PERFORM key_id
	FROM pgcryptokey
	WHERE pgcryptokey.name = v_name AND superseded_by IS NOT NULL;
	IF FOUND THEN
	    RAISE NOTICE 'superseded cryptokey entries for this name exist';
	END IF;

	SELECT key_id INTO v_key_id
	FROM pgcryptokey
	WHERE pgcryptokey.name = v_name AND superseded_by IS NULL;
	IF NOT FOUND THEN
	    RAISE EXCEPTION 'cryptokey entry not found';
	END IF;

	PERFORM drop_cryptokey(v_key_id);

	RETURN TRUE;
END
$$ LANGUAGE plpgsql;
